/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#ifdef WIN32
    #include <direct.h>
    #ifndef NOGUI
        #include <windows.h>
    #endif
    #define PATH_SLASH  '\\'
#else
    #include <unistd.h>
    #define stricmp strcasecmp
    #define NOGUI
    #define PATH_SLASH  '/'
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include "md5.h"



#define VER             "0.4.1"
#define PATCH           0
#define UNPATCH         1
#define ORIGINAL        UNPATCH
#define ISORIGINAL      "File is already the original"
#define ISPATCHED       "File is already patched"
#define SUCCESS         "File has been successfully patched!\n"             \
                        "\n"                                                \
                        "%d bytes changed\n"                                \
                        "\n"                                                \
                        "Do you want to see the report of the changes?\n"
#define MD5DIFF         "Different MD5 checksum, file is not the original\n"                        \
                        "Want you try to force the patching???\n"                                   \
                        "\n"                                                                        \
                        "The process terminates automatically if finds a byte to modify that is\n"  \
                        "different than the original so you can recover your original file using\n" \
                        "the UNPATCH option of this tool\n"                                         \
                        "\n"                                                                        \
                        "Anyway I highly suggest to make a temporary BACKUP of your file if you\n"  \
                        "decide to choose YES and forcing the patching\n"
#define BYTEDIFF        "A byte in your file is not equal to my original file, I cannot continue\n"     \
                        "If you have not made a backup copy of your file you can recover it using\n"    \
                        "the UNPATCH option (NO button) from the initial menu\n"                        \
                        "\n"                                                                            \
                        "%d bytes changed during the process\n"                                         \
                        "\n"                                                                            \
                        "Do you want to see the report of the changes?\n"
#define MSG             "# Lame patcher "VER"\n"                                \
                        "# by Luigi Auriemma\n"                                 \
                        "# e-mail: aluigi@autistici.org\n"                      \
                        "# web:    aluigi.org\n"                                \
                        "\n"                                                    \
                        "Main folder  \t%s\n"                                   \
                        "Data file    \t%s\n"                                   \
                        "Filename     \t%s\n"                                   \
                        "Filesize     \t%u\n"                                   \
                        "Original MD5 \t%s\n"                                   \
                        "Patched MD5  \t%s\n"                                   \
                        "\n"                                                    \
                        "Usage:\n"                                              \
                        "      YES:    \tPATCH \t\t(original => patched)\n"     \
                        "      NO:     \tUNPATCH \t(restore the original)\n"    \
                        "      CANCEL: \tEXIT\n"
#define REPORTHEAD      "OFFSET    \tFROM \tTO\n"
#define REPORTMSG       "%08X \t%02X \t%02X\n"
                    //  "11223344 \t11 \t22\n" <= output is the same size
                    //  anyway I use sizeof(REPORT) so the length is ever +1


#include "iofuncs.h"



head_t  head;
FILE    *lpatch  = NULL;
int     forced   = 0;
uint8_t *title   = NULL,
        *intro   = NULL,
        *comment = NULL,
        *report  = NULL;



#include "lpatch_newmode.h"



int  lpatch_apply(FILE *patchfile, int patch_unpatch);
void patch(uint8_t *name, int choice);
FILE *lpatch_fopen(uint8_t **fname, uint8_t *mode, uint8_t *notfound_msg);   // %s required!!!
void newmode(void);



int main(int argc, char *argv[]) {
    struct stat xstat;
    int     bytes,
            ans;
    uint8_t md5_original[33],
            md5_patch[33],
            *filename,
            *lpach_file;

    setbuf(stdout, NULL);
    setbuf(stdin, NULL);

    current_folder = getcwd(NULL, 0);

    lpach_file = "lpatch.dat";
    if(argc > 1) lpach_file = argv[1];

    lpatch = lpatch_fopen(
        &lpach_file,
        "rb",
        "# Lame patcher "VER"\n"
        "The file %s doesn't exist in the current directory or can't be read\n"
        "You need this file because it contains all the required informations for\n"
        "patching your files\n"
        "This tool supports both the lpatch.dat and *.lpatch files\n"
        "Do you want to select this or another .dat or .lpatch file manually?\n");

    read_head(lpatch, &head);
    if(!memcmp(&head, "================================", 32)) {
        newmode();
        fclose(lpatch);
        return(0);
    }

    filename = frss(lpatch, NULL, head.namelen);

    title    = frss(lpatch, NULL, head.titlelen);

    if(head.commlen > 0) {
        comment = frss(lpatch, NULL, head.commlen);
    }

    make_md5_string(head.md51, md5_original);
    make_md5_string(head.md52, md5_patch);

    fstat(fileno(lpatch), &xstat);
    bytes = (xstat.st_size - ftell(lpatch)) / sizeof(fix_t);
    report = malloc(sizeof(REPORTHEAD) + (sizeof(REPORTMSG) * bytes));
    if(!report) std_err(NULL);
    sprintf(report, REPORTHEAD);

    current_folder = getcwd(NULL, 0);

    ans = msgbox(
        MB_YESNOCANCEL | MB_ICONINFORMATION,
        title,
        MSG,
            current_folder,
            lpach_file,
            filename,
            head.size,
            md5_original,
            md5_patch);

    switch(ans) {
        case IDYES:     patch(filename, PATCH);     break;
        case IDNO:      patch(filename, UNPATCH);   break;
        case IDCANCEL:                              break;
        default:                                    break;
    }

    fclose(lpatch);
    free(filename);
    free(title);
    if(comment) free(comment);
    if(report)  free(report);
    return(0);
}



int lpatch_apply(FILE *patchfile, int patch_unpatch) {
    fix_t   fix;
    int     crc_err    = 0,
            t,
            ans,
            changed    = 0,
            byte_patch = 0;
    uint8_t *hash,
            *x1,
            *x2,
            *rep;

    hash = calc_md5(patchfile);

    if(patch_unpatch == PATCH) {
        if(memcmp(hash, head.md51, 16)) crc_err = 1;
        x1 = &fix.original;
        x2 = &fix.patch;
    } else {
        if(memcmp(hash, head.md52, 16)) crc_err = 1;
        x1 = &fix.patch;
        x2 = &fix.original;
    }

    if(crc_err) {
        if(!memcmp(hash, head.md51, 16)) {
            std_err(ISORIGINAL);
        } else if(!memcmp(hash, head.md52, 16)) {
            std_err(ISPATCHED);
        } else {
            ans = msgbox(
                MB_YESNO | MB_ICONERROR,
                "Error",
                MD5DIFF);
            if(ans == IDNO) exit(1);
            forced = 1;
        }
    }

    rep = report + strlen(report);
    for(;;) {
        if(fgetc(lpatch) < 0) break;    // feof(lpatch) doesn't work!!!
        fseek(lpatch, -1, SEEK_CUR);    // so I must use this work-around

        read_fix(lpatch, &fix);

        if(fseek(patchfile, fix.offset, SEEK_SET)
          < 0) std_err(NULL);

        if(forced) {
            t = fgetc(patchfile);
            if(t < 0) std_err("wrong file offset");
            if(t == *x2) {
                byte_patch++;
            } else if(t != *x1) {
                ans = msgbox(
                    MB_YESNO | MB_ICONERROR,
                    "Alert",
                    BYTEDIFF,
                        changed);
                if(ans == IDYES) {
                    msgbox(
                        MB_ICONINFORMATION,
                        "Report",
                        "%s",
                            report);
                }
                return(-1);
            }
            fseek(patchfile, -1, SEEK_CUR);
        }
        fputc(*x2, patchfile);
        fflush(patchfile);
        changed++;

        fputc('.', stdout);
        rep += sprintf(rep, REPORTMSG, fix.offset, *x1, *x2);
    }

    if(forced && (byte_patch == changed)) {
        std_err((patch_unpatch == PATCH) ? ISPATCHED : ISORIGINAL);
    }

    return(changed);
}



void patch(uint8_t *name, int choice) {
    FILE    *patchfile;
    int     ans,
            changed;

    patchfile = lpatch_fopen(
        &name,
        "r+b",
        "The file %s has not been found in the current directory\n"
        "Do you want to select this file manually?");

    changed = lpatch_apply(patchfile, choice);

    fclose(patchfile);

    if(changed < 0) return;

    ans = msgbox(
        MB_YESNO | MB_ICONINFORMATION,
        "Success",
        SUCCESS,
            changed);

    if(ans == IDYES) {
        msgbox(
            MB_ICONINFORMATION,
            "Report",
            "%s",
                report);
    }

    if(comment) {
        msgbox(
            MB_ICONINFORMATION,
            "Comment",
            "%s",
                comment);
    }
}



FILE *lpatch_fopen(uint8_t **fname, uint8_t *mode, uint8_t *notfound_msg) {
    int     ans;
    FILE    *fd;
    uint8_t *name = NULL;

    if(fname) name = *fname;

    if(name) {
        fd = fopen(name, mode);
        if(fd) return(fd);

        ans = msgbox(
            MB_YESNO | MB_ICONINFORMATION,
            "File not found",
            notfound_msg,
                name);
        if(ans == IDNO) exit(1);
    }

    name = get_file(NULL);
    fd = fopen(name, mode);
    if(!fd) std_err(NULL);
    if(fname) *fname = name;
    return(fd);
}


