/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

enum {
    CMD_TITLE,
    CMD_INTRO,
    CMD_MD5,
    CMD_FILE,
    CMD_RVA,
    CMD_COMMENT,
    CMD_OFFSET,
    CMD_ONLY_ONE,
    CMD_STRING,
    CMD_BYTES_ORIGINAL,
    CMD_BYTES_PATCH,
    CMD_NONE = -1
};



    /* NEWMODE_NUM is studied for 8 bit numbers (1 byte */
#define NEWMODE_NUM_ADD         0x0100
#define NEWMODE_NUM_SUB         0x0200
#define NEWMODE_TYPE_SKIP       0x0400
#define NEWMODE_TYPE_32_OFFSET  0x0800
#define NEWMODE_TYPE_32         0x1000



FILE        *fd_newmode           = NULL;
fix_t       **offbyte             = NULL;
uint32_t    fd_newmode_size       = 0,
            rva                   = 0;
int         offbyte_num           = 0,
            offbyte_alloc_size    = 0,
            newmode_original_size = 0,
            newmode_patch_size    = 0,
            only_one              = 0,
            newmode_md5_hash_num  = 0;
int16_t     *newmode_original     = NULL,   /* int16 is needed for dynamic bytes  */
            *newmode_patch        = NULL,
            newmode_num_type      = 0;
uint8_t     *newmode_fname        = NULL,
            **newmode_md5_hash    = NULL,
            *fd_newmode_md5       = NULL;



void newmode_backup(void) {
    FILE    *fd;
    int     len;
    uint8_t *buff,
            *bckname;

    bckname = malloc(strlen(newmode_fname) + 32);
    if(!bckname) std_err(NULL);
    sprintf(bckname, "%s.LPATCH_BACKUP", newmode_fname);

    buff = malloc(BUFFSZ);
    if(!buff) std_err(NULL);

    fd = fopen(bckname, "wb");
    if(!fd) std_err(NULL);
    rewind(fd_newmode);
    while((len = fread(buff, 1, BUFFSZ, fd_newmode))) {
        fwrite(buff, len, 1, fd);
    }
    fclose(fd);

    free(buff);
    free(bckname);
}



uint32_t get_num(uint8_t *data) {
    uint32_t    num = 0;
    uint8_t     *p;

    newmode_num_type = 0;
    if(*data == '+') { newmode_num_type |= NEWMODE_NUM_ADD;         data++; }
    if(*data == '-') { newmode_num_type |= NEWMODE_NUM_SUB;         data++; }
    if(*data == '*') { newmode_num_type |= NEWMODE_TYPE_SKIP;       data++; }
    if(*data == '^') { newmode_num_type |= NEWMODE_TYPE_32_OFFSET;  data++; }
    if(*data == '|') { newmode_num_type |= NEWMODE_TYPE_32;         data++; }

    if(!data[0])        return(0);
    if(data[0] == '?')  return(-1);     // dynamic byte

    if(data[0] == '\'') {
        if(data[1] == '\\') {           // \n and so on
            for(p = data; *p; p++) *p = tolower(*p);
            switch(data[2]) {
                case '\"':  num = '\"';                     break;
                case '\'':  num = '\'';                     break;
                case '\\':  num = '\\';                     break;
                case 'a':   num = '\a';                     break;
                case 'b':   num = '\b';                     break;
                case '\f':  num = '\f';                     break;
                case '\n':  num = '\n';                     break;
                case '\r':  num = '\r';                     break;
                case '\t':  num = '\t';                     break;
                case '\v':  num = '\v';                     break;
                case 'x':   sscanf(data + 3, "%x", &num);   break;  // hex
                default:    sscanf(data + 2, "%o", &num);   break;  // octal
            }
        } else {
            num = data[1];              // 'a'
        }
    } else {
        for(p = data; *p; p++) *p = tolower(*p);
        if((strlen(data) > 2) && (data[0] == '0') && (data[1] == 'x')) {
            data += 2;
        } else if((strlen(data) > 1) && (data[0] == '$')) {
            data++;
        }
        sscanf(data, "%x", &num);
    }

    if(newmode_num_type & NEWMODE_NUM_ADD) num |= NEWMODE_NUM_ADD;
    if(newmode_num_type & NEWMODE_NUM_SUB) num |= NEWMODE_NUM_SUB;
    return(num);
}



void add_offset(uint32_t offset, int original, int patch) {
    int     i;

    if(patch & NEWMODE_NUM_ADD) patch = original + patch;
    if(patch & NEWMODE_NUM_SUB) patch = original - patch;
    original &= 0xff;
    patch    &= 0xff;

    if(original == patch) return;
    for(i = 0; i < offbyte_num; i++) {
        if(offset == offbyte[i]->offset) return;
    }

    if(!offbyte) {
        offbyte_num        = 0;
        offbyte_alloc_size = 128;
        offbyte            = malloc(sizeof(fix_t *) * offbyte_alloc_size);
        if(!offbyte) std_err(NULL);
    }
    offbyte_num++;
    if(offbyte_num > offbyte_alloc_size) {
        offbyte_alloc_size += 128;
        offbyte             = realloc(offbyte, sizeof(fix_t *) * offbyte_alloc_size);
        if(!offbyte) std_err(NULL);
    }
    offbyte[offbyte_num - 1]           = malloc(sizeof(fix_t));
    if(!offbyte[offbyte_num - 1]) std_err(NULL);
    offbyte[offbyte_num - 1]->offset   = offset;
    offbyte[offbyte_num - 1]->original = original;
    offbyte[offbyte_num - 1]->patch    = patch;
}



uint8_t *get_newmode_cmd(uint8_t *line, int *cmdnum) {
    int         i,
                cmdret;
    uint8_t     *cmd,
                *p,
                *l;
    static const uint8_t *command[] = {
                "TITLE",
                "INTRO",
                "MD5",
                "FILE",
                "RVA",
                "COMMENT",
                "OFFSET",
                "ONLY_ONE",
                "STRING",
                "BYTES_ORIGINAL",
                "BYTES_PATCH",
                NULL };

    cmdret  = CMD_NONE;
    *cmdnum = CMD_NONE;

    l = line + delimit(line);

    for(p = line; *p; p++) {        // clear start
        if((*p != ' ') && (*p != '\t')) break;
    }
    if(!*p) return(NULL);

    cmd = p;                        // cmd

    for(l--; l > p; l--) {          // clear end
        if(*l > ' ') break;
    }
    *(l + 1) = 0;

    if((*cmd == '=') || (*cmd == '#') || (*cmd == '/') || (*cmd == ';')) return(NULL);

    for(p = cmd; *p > ' '; p++);    // find where the command ends

    for(i = 0; command[i]; i++) {
        if(!memcmp(cmd, command[i], p - cmd)) {
            cmdret = i;
            break;
        }
    }

    if(cmdret != CMD_NONE) {        // skip the spaces between the comamnd and the instructions
        for(; *p; p++) {
            if((*p != ' ') && (*p != '\t')) break;
        }
        cmd = p;
    }

    // do not enable this or will not work!
    // if((*cmd == '=') || (*cmd == '#') || (*cmd == '/') || (*cmd == ';')) return("");

    *cmdnum = cmdret;
    return(cmd);
}



    /* here we catch each line (till line feed) */
    /* returns a pointer to the next line       */
uint8_t *get_line(uint8_t *data) {
    uint8_t  *p;

    for(p = data; *p && (*p != '\n') && (*p != '\r'); p++);
    if(!*p) return(NULL);
    *p = 0;
    for(p++; *p && ((*p == '\n') || (*p == '\r')); p++);
    if(!*p) return(NULL);
    return(p);
}



    /* here we catch each element of the line */
    /* returns a pointer to the next element  */
uint8_t *get_element(uint8_t **data) {
    uint8_t  *p;

    p = *data;
    if(*p == '"') {     // string
        p++;
        *data = p;
        while(*p && (*p != '"')) p++;
    } else {
        while(*p && (*p != '\t') && (*p != ' ')) p++;
    }

    if(!*p) return(NULL);
    *p = 0;

    for(p++; *p && ((*p == '\t') || (*p == ' ')); p++);
    if(!*p) return(NULL);
    return(p);
}



int search_file(void) {
    uint32_t    offset,
                poff,
                bit32;
    int         i,
                flen,
                blen,
                difflen,
                changes = 0;
    uint8_t     *buff   = NULL,
                *limit,
                *p,
                *l;

    /*
    printf("newmode_original_size: %d -> %d\n", newmode_original_size, newmode_patch_size);
    */

    if(only_one && offbyte_num) goto quit;

    if(
         (!newmode_original_size || !newmode_patch_size)
      || (newmode_original_size > fd_newmode_size)
      || (newmode_patch_size    > fd_newmode_size)) {
        goto quit;
    }

    rewind(fd_newmode);

    difflen = 0;
    if(newmode_patch_size > newmode_original_size) difflen = newmode_patch_size - newmode_original_size;

    buff = malloc(BUFFSZ + newmode_original_size + difflen);
    if(!buff) std_err(NULL);

    blen   = 0;
    offset = 0;
    flen   = fread(buff, 1, BUFFSZ + newmode_original_size, fd_newmode);

    do {
        // if(flen < newmode_original_size) break;
        limit = buff + flen + blen;                     // limit is just the end of buff
        l     = (limit - newmode_original_size) + 1;    // used later, in short we need to
                                                        // place the original_size-1 buffer
                                                        // at the beginning of buff

        for(p = buff; p < limit; p++) {
            for(i = 0; i < newmode_original_size; i++) {
                if(newmode_original[i] & NEWMODE_TYPE_SKIP) continue;
                if(p[i] != newmode_original[i]) break;
            }
            if(i != newmode_original_size) continue;

            if((p + newmode_patch_size) > limit) {
                if(fread(limit, difflen, 1, fd_newmode) != 1) goto quit;
                fseek(fd_newmode, -difflen, SEEK_CUR);
            }

            poff = offset + (p - buff) - blen;

            for(i = 0; i < newmode_patch_size; i++) {
                if(newmode_patch[i] & NEWMODE_TYPE_SKIP) continue;

                if(newmode_patch[i] & NEWMODE_TYPE_32_OFFSET) {
                    bit32 = ((newmode_patch[i    ] & 0xff)      )
                          | ((newmode_patch[i + 1] & 0xff) <<  8)
                          | ((newmode_patch[i + 2] & 0xff) << 16)
                          | ((newmode_patch[i + 3] & 0xff) << 24);

                    bit32 -= ((poff + i + rva) + 4);    /* absolute offset */

                            add_offset(poff + i, p[i], (bit32      ) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >>  8) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >> 16) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >> 24) & 0xff);

                } else if(newmode_patch[i] & NEWMODE_TYPE_32) {
                    bit32 = ((newmode_patch[i    ] & 0xff)      )
                          | ((newmode_patch[i + 1] & 0xff) <<  8)
                          | ((newmode_patch[i + 2] & 0xff) << 16)
                          | ((newmode_patch[i + 3] & 0xff) << 24);

                            add_offset(poff + i, p[i], (bit32      ) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >>  8) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >> 16) & 0xff);
                    ++i;    add_offset(poff + i, p[i], (bit32 >> 24) & 0xff);

                } else {
                    add_offset(poff + i, p[i], newmode_patch[i]);
                }
            }
            changes++;
            p += newmode_original_size - 1; // -1 because we already increment p each time
            /* would be useful to increment p for the size of newmode_patch_size or not? */
        }

        blen = limit - l;           // blen is usually newmode_original_size-1
        memmove(buff, l, blen);     // copy the unread data at the beginning of buff
        offset += flen;             // set the correct offset
        flen = fread(buff + blen, 1, BUFFSZ + newmode_original_size - blen, fd_newmode);
    } while(flen);

quit:
    if(buff) free(buff);

    free(newmode_original);
    newmode_original      = NULL;
    newmode_original_size = 0;

    free(newmode_patch);
    newmode_patch         = NULL;
    newmode_patch_size    = 0;

    return(changes);
}



int16_t *create_newmode_string(uint8_t *in, int *outlen) {
    int         i,
                inlen;
    int16_t     *out;

    inlen   = strlen(in);
    *outlen = inlen;

    out = malloc(sizeof(int16_t) * inlen);
    if(!out) std_err(NULL);
    for(i = 0; i < inlen; i++) {
        out[i] = in[i];
        if(out[i] == '?') out[i] = NEWMODE_TYPE_SKIP;
    }
    return(out);
}



void newmode_apply_patch(void) {
    int         i,
                ans;
    uint8_t     *rep;

    if(!fd_newmode) return;
    if(!offbyte_num) {
        msgbox(
            MB_ICONINFORMATION,
            "Alert",
            "there are no bytes to change in the file");
        return;
    }

    ans = msgbox(
        MB_YESNO | MB_ICONINFORMATION,
        "Success",
        "I'm ready for starting the patching of the file\n"
        "Do you want to create a backup file (%s.LPATCH_BACKUP)?",
            newmode_fname);
    if(ans == IDYES) newmode_backup();

    report = malloc(sizeof(REPORTHEAD) + (sizeof(REPORTMSG) * offbyte_num));
    if(!report) std_err(NULL);
    rep = report +
        sprintf(report, REPORTHEAD);

    for(i = 0; i < offbyte_num; i++) {
        fseek(fd_newmode, offbyte[i]->offset, SEEK_SET);
        fputc(offbyte[i]->patch, fd_newmode);
        rep += sprintf(rep, REPORTMSG, offbyte[i]->offset, offbyte[i]->original, offbyte[i]->patch);
        free(offbyte[i]);
    }
    free(offbyte);
    offbyte = NULL;
    fflush(fd_newmode);

    ans = msgbox(
        MB_YESNO | MB_ICONINFORMATION,
        "Success",
        SUCCESS,
            offbyte_num);
    offbyte_num = 0;

    if(ans == IDYES) {
        msgbox(
            MB_ICONINFORMATION,
            "Report",
            "%s",
                report);
    }
    free(report);
    report = NULL;

    if(comment) {
        msgbox(
            MB_ICONINFORMATION,
            "Comment",
            "%s",
                comment);
        free(comment);
        comment = NULL;
    }
}



void newmode_title(uint8_t *line) {
    if(title) free(title);      // remove duplicates
    title = strdup(line);
}



void newmode_intro(uint8_t *line) {
    if(intro) free(intro);      // remove duplicates
    intro = strdup(line);
}



void newmode_md5(uint8_t *line) {
    int     i,
            c;
    uint8_t *next;

    next = NULL;
    do {
        next = get_line(line);

        newmode_md5_hash = realloc(newmode_md5_hash, sizeof(uint8_t *) * (newmode_md5_hash_num + 1));
        if(!newmode_md5_hash) std_err(NULL);
        newmode_md5_hash[newmode_md5_hash_num] = malloc(16);

        for(i = 0; i < 16; i++) {
            while(line[0] <= ' ') line++;
            line[0] = tolower(line[0]);
            line[1] = tolower(line[1]);
            if(!line[0] || !line[1]) {
                free(newmode_md5_hash[newmode_md5_hash_num]);
                return;
            }
            sscanf(line, "%02x", &c);
            line += 2;
            newmode_md5_hash[newmode_md5_hash_num][i] = c;
        }
        newmode_md5_hash_num++;

        line = next;
    } while(next);
}



void newmode_file(uint8_t *line) {
    struct stat xstat;
    int     i;

    get_line(line);             /* initialize line */

    if(fd_newmode) {
        newmode_apply_patch();  /* it's time to apply the patch */
        fclose(fd_newmode);
        fd_newmode      = NULL;
        fd_newmode_size = 0;
    }
    rva            = 0;
    only_one       = 0;
    fd_newmode_md5 = NULL;

    if(title) {
        msgbox(
            MB_ICONINFORMATION,
            "Lame patcher "VER,
            "%s",
                title);
        free(title);
        title = NULL;
    }
    if(intro) {
        msgbox(
            MB_ICONINFORMATION,
            "Lame patcher "VER,
            "%s",
                intro);
        free(intro);
        intro = NULL;
    }

    newmode_fname = get_file(line);

    fd_newmode = fopen(newmode_fname, "r+b");
    if(!fd_newmode) std_err(NULL);

    fstat(fileno(fd_newmode), &xstat);
    fd_newmode_size = xstat.st_size;

    if(newmode_md5_hash) {
        fd_newmode_md5 = calc_md5(fd_newmode);

        for(i = 0; i < newmode_md5_hash_num; i++) {
            if(!memcmp(fd_newmode_md5, newmode_md5_hash[i], 16)) break;
        }

        if(i == newmode_md5_hash_num) {
            std_err("the MD5 hash specified in the data file differs from that of your file");
        }
        newmode_md5_hash     = NULL;
        newmode_md5_hash_num = 0;
    }
}



void newmode_rva(uint8_t *line) {
    rva = get_num(line);
}



void newmode_comment(uint8_t *line) {
    if(comment) free(comment);  // remove duplicates
    comment = strdup(line);
}



void newmode_offset(uint8_t *line) {
    uint32_t    offset,
                old_offset = 0,
                bit32      = 0;
    int         x1,
                x2,
                fc;
    uint8_t     *next,
                *soffset,
                *sx1,
                *sx2;

    if(!fd_newmode) std_err("the file to patch (FILE) has not been opened or specified yet, check the data file");

    next = NULL;
    do {
        next = get_line(line);          /* get the current line */

        soffset = line;                 /* read all the elements */
        sx1     = get_element(&soffset);    if(!sx1) return;
        sx2     = get_element(&sx1);        if(!sx2) return;
                  get_element(&sx2);

                                        /* offset */
        if(soffset[0] == '+') {         /* next offsets */
            if(!soffset[1] || (soffset[1] == '+')) {
                offset = old_offset + 1;
            } else {
                offset = old_offset + get_num(soffset + 1);
            }
        } else if(soffset[0] == '-') {  /* previous offsets */
            if(!soffset[1] || (soffset[1] == '-')) {
                offset = old_offset - 1;
            } else {
                offset = old_offset - get_num(soffset + 1);
            }
        } else {
            offset = get_num(soffset);
        }

        x1      = get_num(sx1);         /* original byte */

        x2      = get_num(sx2);         /* patched byte */

        if(newmode_num_type & NEWMODE_TYPE_32_OFFSET) bit32 = x2 - (offset + 4);
        if(newmode_num_type & NEWMODE_TYPE_32)        bit32 = x2;

        offset -= rva;                  /* convert in file offset */

        if(offset > fd_newmode_size) {
            std_err("the offset specified in the data file cannot be reached in your input file, probably smaller");
        }

        fseek(fd_newmode, offset, SEEK_SET);
        fc = fgetc(fd_newmode);
        if(fc < 0) std_err("the offset specified in the data file cannot be reached in your input file, probably smaller");

        if((x1 >= 0) && (fc != x1)) {
            std_err("one or more bytes in the file to patch are not the originals, probably your file is not the same of the author of this patch");
        }

        if(bit32) {
                                        add_offset(offset++, x1, (bit32      ) & 0xff);
            x1 = fgetc(fd_newmode);     add_offset(offset++, x1, (bit32 >>  8) & 0xff);
            x1 = fgetc(fd_newmode);     add_offset(offset++, x1, (bit32 >> 16) & 0xff);
            x1 = fgetc(fd_newmode);     add_offset(offset,   x1, (bit32 >> 24) & 0xff);
            bit32 = 0;
        } else {
            add_offset(offset, x1, x2);
        }

        old_offset = offset + rva;

        line = next;
    } while(next);
}


void newmode_only_one(uint8_t *line) {
    only_one = 1;
}



void newmode_string(uint8_t *line) {
    uint8_t     *next,
                *original,
                *patch;

    if(!fd_newmode) std_err("the file to patch (FILE) has not been opened or specified yet, check the data file");

    next = NULL;
    do {
        next = get_line(line);

        original = line;
        patch    = get_element(&original);  if(!patch) return;
                   get_element(&patch);

        newmode_original = create_newmode_string(original, &newmode_original_size);
        newmode_patch    = create_newmode_string(patch,    &newmode_patch_size);

        if(!search_file()) {
            // msgbox(
            //     MB_ICONERROR,
            //     "Alert",
            //     "there are no bytes to change in the file, I continue with the next patch");
            return;
        }

        line = next;
    } while(next);
}



void newmode_bytes(uint8_t *line, int original_patch) {
#define ADDBYTE(A)  oplen++;                                    \
                    op = realloc(op, sizeof(int16_t) * oplen);  \
                    if(!op) std_err(NULL);                      \
                    op[oplen - 1] = byte | A;

    int         i,
                byte,
                oplen = 0;
    int16_t     *op   = NULL;
    uint8_t     *next,
                *sc,
                *scn;

    if(!fd_newmode) std_err("the file to patch (FILE) has not been opened or specified yet, check the data file");

    next = NULL;
    do {
        next = get_line(line);

        sc  = line;
        scn = NULL;
        do {
            scn = get_element(&sc);

            if((sc[0] == '#') || (sc[0] == '/') || (sc[0] == ';')) break; // comments
            byte = get_num(sc);

            if(newmode_num_type & NEWMODE_TYPE_SKIP) {
                oplen += byte;
                op = realloc(op, sizeof(int16_t) * oplen);
                if(!op) std_err(NULL);
                for(i = oplen - byte; i < oplen; i++) op[i] = NEWMODE_TYPE_SKIP;

            } else if(newmode_num_type & NEWMODE_TYPE_32_OFFSET) {
                oplen += 4;
                op = realloc(op, sizeof(int16_t) * oplen);
                if(!op) std_err(NULL);
                op[oplen - 4] = ((byte      ) & 0xff) | NEWMODE_TYPE_32_OFFSET;
                op[oplen - 3] = ((byte >>  8) & 0xff) | NEWMODE_TYPE_32_OFFSET;
                op[oplen - 2] = ((byte >> 16) & 0xff) | NEWMODE_TYPE_32_OFFSET;
                op[oplen - 1] = ((byte >> 24) & 0xff) | NEWMODE_TYPE_32_OFFSET;

            } else if(newmode_num_type & NEWMODE_TYPE_32) {
                oplen += 4;
                op = realloc(op, sizeof(int16_t) * oplen);
                if(!op) std_err(NULL);
                op[oplen - 4] = ((byte      ) & 0xff) | NEWMODE_TYPE_32;
                op[oplen - 3] = ((byte >>  8) & 0xff) | NEWMODE_TYPE_32;
                op[oplen - 2] = ((byte >> 16) & 0xff) | NEWMODE_TYPE_32;
                op[oplen - 1] = ((byte >> 24) & 0xff) | NEWMODE_TYPE_32;

            } else {
                oplen++;
                op = realloc(op, sizeof(int16_t) * oplen);
                if(!op) std_err(NULL);
                op[oplen - 1] = byte;
            }

            sc = scn;
        } while(scn);

        line = next;
    } while(next);

    if(original_patch == ORIGINAL) {
        newmode_original      = op;
        newmode_original_size = oplen;

    } else {
        newmode_patch         = op;
        newmode_patch_size    = oplen;

        if(!search_file()) {
            // msgbox(
            //     MB_ICONERROR,
            //     "Alert",
            //     "there are no bytes to change in the file, I continue with the next patch");
            return;
        }
    }

#undef ADDBYTE
}



void newmode_cmd(int cmdnum, uint8_t *line) {
    switch(cmdnum) {
        case CMD_TITLE:             newmode_title(line);            break;
        case CMD_INTRO:             newmode_intro(line);            break;
        case CMD_MD5:               newmode_md5(line);              break;
        case CMD_FILE:              newmode_file(line);             break;
        case CMD_RVA:               newmode_rva(line);              break;
        case CMD_COMMENT:           newmode_comment(line);          break;
        case CMD_OFFSET:            newmode_offset(line);           break;
        case CMD_ONLY_ONE:          newmode_only_one(line);         break;
        case CMD_STRING:            newmode_string(line);           break;
        case CMD_BYTES_ORIGINAL:    newmode_bytes(line, ORIGINAL);  break;
        case CMD_BYTES_PATCH:       newmode_bytes(line, PATCH);     break;
        default:                                                    break;
    }
}



void newmode(void) {
    int     len,
            currlen,
            bufflen,
            oldnum,
            cmdnum,
            tmp;
    uint8_t line[1024],
            *buff,
            *buff_limit,
            *data,
            *ins;

    rewind(lpatch);

    bufflen    = 1024;
    buff       = malloc(bufflen);
    if(!buff) std_err(NULL);
    data       = buff;
    buff_limit = buff + bufflen;
    buff[0]    = 0;
    line[0]    = 0;
    oldnum     = CMD_NONE;

    while(fgets(line, sizeof(line), lpatch)) {
        ins = get_newmode_cmd(line, &cmdnum);
        if(!ins) continue;

        if(oldnum == CMD_NONE) oldnum = cmdnum;
        if(cmdnum == CMD_NONE) cmdnum = oldnum;
        if(cmdnum != oldnum) {
            tmp    = cmdnum;
            cmdnum = oldnum;
            oldnum = tmp;

            newmode_cmd(cmdnum, buff);

            data = buff;
        }

        len = strlen(ins);  // allocation
        if((data + len) >= buff_limit) {
            currlen    = data - buff;
            bufflen    = currlen + 1 + len + 1; // 1 for \n and 1 for the final NULL byte
            buff       = realloc(buff, bufflen);
            if(!buff) std_err(NULL);
            data       = buff + currlen;
            buff_limit = buff + bufflen;
        }

        if(data > buff) data += sprintf(data, "\n");
        data += sprintf(data, "%s", ins);
        line[0] = 0;
    }
        // the remaining line
    cmdnum = oldnum;
    if((cmdnum != CMD_NONE) && (data != buff)) newmode_cmd(cmdnum, buff);

    newmode_apply_patch();

    free(buff);
}


